#include <platform.h>
#include "motor_control.h"

out port pwm_out     = on tile[2]:XS1_PORT_4A;
out port dir_ctrl    = on tile[2]:XS1_PORT_4B;
out port not_standby = on tile[2]:XS1_PORT_1A;

#define REVERSE 0x0A
#define FORWARD 0x05
#define MIN_DUTY_CYCLE 0

void pwm(chanend from_controller, chanend to_controller) {
    const unsigned RESOLUTION = 20;
    const unsigned STEPS = 255;

    timer tmr;
    unsigned t;
    int output = 0;
    int dirty_flag;
    int hi[LINES] = {0, 0};
    int cmd;

    dir_ctrl <: FORWARD;
    not_standby <: 1;

    tmr :> t;
    output = 0;
    while(1) {
        tmr when timerafter(t) :> void;
        pwm_out <: 0xFF;
        tmr when timerafter(t+RESOLUTION * hi[0]) :> void;
        pwm_out <: 0x0 ;
        t += RESOLUTION * STEPS;

        // get current motor_cmd (request/response)
        to_controller <: 0;
        from_controller :> cmd;
        if (cmd < 0) {
            dir_ctrl <: REVERSE;
            cmd = -cmd;
        } else {
            dir_ctrl <: FORWARD;
        }
        for(int i=0; i < LINES; i++) {
            if (cmd < MIN_DUTY_CYCLE) cmd = 0;
            hi[i] = cmd;
        }

    }
    while(1) {
        // start of PWM cycle
        output = 0xFF;
        dirty_flag = 1;  // this will cause output to start high (assuming pwm not set to 0)
        for(int step=0; step < STEPS; step++) {
            for (int i=0; i < LINES; i++) {
                if (hi[i] == step) {
                    dirty_flag = 1;
                    output &= ~(1 << i);
                }
            }
            if(dirty_flag) {
                //tmr when timerafter(t + RESOLUTION * step) :> void;
                tmr when timerafter(t) :> void;
                pwm_out <: output;
                dirty_flag = 0;
            }
            t += RESOLUTION;
        }

        // get current motor_cmd (request/response)
        to_controller <: 0;
        from_controller :> cmd;
        if (cmd < 0) {
            dir_ctrl <: REVERSE;
            cmd = -cmd;
        } else {
            dir_ctrl <: FORWARD;
        }
        for(int i=0; i < LINES; i++) {
            if (cmd < MIN_DUTY_CYCLE) cmd = 0;
            hi[i] = cmd;
        }
       //t += RESOLUTION * STEPS;
    }
}
