/*
 * mpu6050.c
 *
 *  Created on: Nov 28, 2012
 *      Author: ah6
 */
#include <math.h>
#include "mpu6050.h"

#define GYRO_CONTRIBUTION 0.995

// this values derive from gyro rate (we're setting to
// 500 dps and word width (16 bits).  Multiplying by
// 10 to avoid floating point


void delay_ms(const int delay) {
    timer tmr;
    unsigned int t;
    tmr :> t;
    tmr when timerafter(t+delay*100000) :> void;
}

void mpu_write(struct r_i2c &mpu, const int reg, const unsigned char value) {
    unsigned char data[1];
    data[0] = value;
    i2c_master_write_reg(MPU6050_ADDRESS, reg, data, 1, mpu);
}

unsigned char mpu_read_byte(struct r_i2c &mpu, const int reg) {
    unsigned char data[1];
    i2c_master_read_reg(MPU6050_ADDRESS, reg, data, 1, mpu);
    return data[0];
}

short mpu_read_short(struct r_i2c &mpu, const int reg) {
    //unsigned char data[2];
    //i2c_master_read_reg(MPU6050_ADDRESS, reg, data, 2, mpu);
    //return (data[0] << 8) | data[1];
    unsigned char hi = mpu_read_byte(mpu, reg);
    unsigned char lo = mpu_read_byte(mpu, reg+1);
    return  hi << 8 | lo;
}

void imu_setup(struct IMU &imu)
{
    //struct r_i2c mpu;

    i2c_master_init(imu.i2c);

    //Sets sample rate to 8000/1+7 = 1000Hz
    mpu_write(imu.i2c, MPU6050_RA_SMPLRT_DIV, 0x07);

    //Disable FSync, 256Hz DLPF
    mpu_write(imu.i2c, MPU6050_RA_CONFIG, 0x00);

    //Disable gyro self tests, scale of 500 degrees/s
    mpu_write(imu.i2c, MPU6050_RA_GYRO_CONFIG, 0x08);

    //Disable accel self tests, scale of +-2g, no DHPF
    mpu_write(imu.i2c, MPU6050_RA_ACCEL_CONFIG, 0x00);

    //Freefall threshold of |0mg|
    mpu_write(imu.i2c, MPU6050_RA_FF_THR, 0x00);

    //Freefall duration limit of 0
    mpu_write(imu.i2c, MPU6050_RA_FF_DUR, 0x00);

    //Motion threshold of 0mg
    mpu_write(imu.i2c, MPU6050_RA_MOT_THR, 0x00);

    //Motion duration of 0s
    mpu_write(imu.i2c, MPU6050_RA_MOT_DUR, 0x00);

    //Zero motion threshold
    mpu_write(imu.i2c, MPU6050_RA_ZRMOT_THR, 0x00);

    //Zero motion duration threshold
    mpu_write(imu.i2c, MPU6050_RA_ZRMOT_DUR, 0x00);

    //Disable sensor output to FIFO buffer
    mpu_write(imu.i2c, MPU6050_RA_FIFO_EN, 0x00);

    //AUX I2C setup
    //Sets AUX I2C to single master control, plus other config
    mpu_write(imu.i2c, MPU6050_RA_I2C_MST_CTRL, 0x00);

    //Setup AUX I2C slaves
    mpu_write(imu.i2c, MPU6050_RA_I2C_SLV0_ADDR, 0x00);
    mpu_write(imu.i2c, MPU6050_RA_I2C_SLV0_REG,  0x00);
    mpu_write(imu.i2c, MPU6050_RA_I2C_SLV0_CTRL, 0x00);
    mpu_write(imu.i2c, MPU6050_RA_I2C_SLV1_ADDR, 0x00);
    mpu_write(imu.i2c, MPU6050_RA_I2C_SLV1_REG,  0x00);
    mpu_write(imu.i2c, MPU6050_RA_I2C_SLV1_CTRL, 0x00);
    mpu_write(imu.i2c, MPU6050_RA_I2C_SLV2_ADDR, 0x00);
    mpu_write(imu.i2c, MPU6050_RA_I2C_SLV2_REG,  0x00);
    mpu_write(imu.i2c, MPU6050_RA_I2C_SLV2_CTRL, 0x00);
    mpu_write(imu.i2c, MPU6050_RA_I2C_SLV3_ADDR, 0x00);
    mpu_write(imu.i2c, MPU6050_RA_I2C_SLV3_REG,  0x00);
    mpu_write(imu.i2c, MPU6050_RA_I2C_SLV3_CTRL, 0x00);
    mpu_write(imu.i2c, MPU6050_RA_I2C_SLV4_ADDR, 0x00);
    mpu_write(imu.i2c, MPU6050_RA_I2C_SLV4_REG,  0x00);
    mpu_write(imu.i2c, MPU6050_RA_I2C_SLV4_DO,   0x00);
    mpu_write(imu.i2c, MPU6050_RA_I2C_SLV4_CTRL, 0x00);
    mpu_write(imu.i2c, MPU6050_RA_I2C_SLV4_DI,   0x00);

    //MPU6050_RA_I2C_MST_STATUS       //Read-only
    //Setup INT pin and AUX I2C pass through
    mpu_write(imu.i2c, MPU6050_RA_INT_PIN_CFG, 0x00);
    //Enable data ready interrupt
    mpu_write(imu.i2c, MPU6050_RA_INT_ENABLE, 0x00);

    //MPU6050_RA_DMP_INT_STATUS       //Read-only
    //MPU6050_RA_INT_STATUS 3A        //Read-only
    //MPU6050_RA_ACCEL_XOUT_H         //Read-only
    //MPU6050_RA_ACCEL_XOUT_L         //Read-only
    //MPU6050_RA_ACCEL_YOUT_H         //Read-only
    //MPU6050_RA_ACCEL_YOUT_L         //Read-only
    //MPU6050_RA_ACCEL_ZOUT_H         //Read-only
    //MPU6050_RA_ACCEL_ZOUT_L         //Read-only
    //MPU6050_RA_TEMP_OUT_H           //Read-only
    //MPU6050_RA_TEMP_OUT_L           //Read-only
    //MPU6050_RA_GYRO_XOUT_H          //Read-only
    //MPU6050_RA_GYRO_XOUT_L          //Read-only
    //MPU6050_RA_GYRO_YOUT_H          //Read-only
    //MPU6050_RA_GYRO_YOUT_L          //Read-only
    //MPU6050_RA_GYRO_ZOUT_H          //Read-only
    //MPU6050_RA_GYRO_ZOUT_L          //Read-only
    //MPU6050_RA_EXT_SENS_DATA_00     //Read-only
    //MPU6050_RA_EXT_SENS_DATA_01     //Read-only
    //MPU6050_RA_EXT_SENS_DATA_02     //Read-only
    //MPU6050_RA_EXT_SENS_DATA_03     //Read-only
    //MPU6050_RA_EXT_SENS_DATA_04     //Read-only
    //MPU6050_RA_EXT_SENS_DATA_05     //Read-only
    //MPU6050_RA_EXT_SENS_DATA_06     //Read-only
    //MPU6050_RA_EXT_SENS_DATA_07     //Read-only
    //MPU6050_RA_EXT_SENS_DATA_08     //Read-only
    //MPU6050_RA_EXT_SENS_DATA_09     //Read-only
    //MPU6050_RA_EXT_SENS_DATA_10     //Read-only
    //MPU6050_RA_EXT_SENS_DATA_11     //Read-only
    //MPU6050_RA_EXT_SENS_DATA_12     //Read-only
    //MPU6050_RA_EXT_SENS_DATA_13     //Read-only
    //MPU6050_RA_EXT_SENS_DATA_14     //Read-only
    //MPU6050_RA_EXT_SENS_DATA_15     //Read-only
    //MPU6050_RA_EXT_SENS_DATA_16     //Read-only
    //MPU6050_RA_EXT_SENS_DATA_17     //Read-only
    //MPU6050_RA_EXT_SENS_DATA_18     //Read-only
    //MPU6050_RA_EXT_SENS_DATA_19     //Read-only
    //MPU6050_RA_EXT_SENS_DATA_20     //Read-only
    //MPU6050_RA_EXT_SENS_DATA_21     //Read-only
    //MPU6050_RA_EXT_SENS_DATA_22     //Read-only
    //MPU6050_RA_EXT_SENS_DATA_23     //Read-only
    //MPU6050_RA_MOT_DETECT_STATUS    //Read-only

    //Slave out, dont care
    mpu_write(imu.i2c, MPU6050_RA_I2C_SLV0_DO, 0x00);
    mpu_write(imu.i2c, MPU6050_RA_I2C_SLV1_DO, 0x00);
    mpu_write(imu.i2c, MPU6050_RA_I2C_SLV2_DO, 0x00);
    mpu_write(imu.i2c, MPU6050_RA_I2C_SLV3_DO, 0x00);
    //More slave config
    mpu_write(imu.i2c, MPU6050_RA_I2C_MST_DELAY_CTRL, 0x00);
    //Reset sensor signal paths
    mpu_write(imu.i2c, MPU6050_RA_SIGNAL_PATH_RESET, 0x00);
    //Motion detection control
    mpu_write(imu.i2c, MPU6050_RA_MOT_DETECT_CTRL, 0x00);
    //Disables FIFO, AUX I2C, FIFO and I2C reset bits to 0
    mpu_write(imu.i2c, MPU6050_RA_USER_CTRL, 0x00);
    //Sets clock source to gyro reference w/ PLL
    mpu_write(imu.i2c, MPU6050_RA_PWR_MGMT_1, 0b00000010);
    //Controls frequency of wakeups in accel low power mode plus the sensor standby modes
    mpu_write(imu.i2c, MPU6050_RA_PWR_MGMT_2, 0x00);
    //MPU6050_RA_BANK_SEL               //Not in datasheet
    //MPU6050_RA_MEM_START_ADDR         //Not in datasheet
    //MPU6050_RA_MEM_R_W                //Not in datasheet
    //MPU6050_RA_DMP_CFG_1              //Not in datasheet
    //MPU6050_RA_DMP_CFG_2              //Not in datasheet
    //MPU6050_RA_FIFO_COUNTH            //Read-only
    //MPU6050_RA_FIFO_COUNTL            //Read-only
    //Data transfer to and from the FIFO buffer
    mpu_write(imu.i2c, MPU6050_RA_FIFO_R_W, 0x00);
    //MPU6050_RA_WHO_AM_I             //Read-only, I2C address

}

{int, int, int} get_gyro_raw(struct IMU &imu) {
    int x = mpu_read_short(imu.i2c, MPU6050_RA_GYRO_XOUT_H);
    int y = mpu_read_short(imu.i2c, MPU6050_RA_GYRO_YOUT_H);
    int z = mpu_read_short(imu.i2c, MPU6050_RA_GYRO_ZOUT_H);

    return {x, y, z};
}

{float, float, float} get_gyro(struct IMU &imu) {
    float gx, gy, gz ;
    int raw_x, raw_y, raw_z ;
    {raw_x, raw_y, raw_z} = get_gyro_raw(imu);

    gx = (raw_x - imu.gyro_x_offset) / GYRO_SENSITIVITY;
    gy = (raw_y - imu.gyro_y_offset) / GYRO_SENSITIVITY;
    gz = (raw_z - imu.gyro_z_offset) / GYRO_SENSITIVITY;

    return {gx, gy, gz};

}

void calibrate_gyros(struct IMU &imu) {
    const int LOOPS = 1000;

    int GYRO_XOUT_OFFSET_SUM = 0;
    int GYRO_YOUT_OFFSET_SUM = 0;
    int GYRO_ZOUT_OFFSET_SUM = 0;
    for(int i = 0; i < LOOPS; i++)
    {
        GYRO_XOUT_OFFSET_SUM += mpu_read_short(imu.i2c, MPU6050_RA_GYRO_XOUT_H);
        GYRO_YOUT_OFFSET_SUM += mpu_read_short(imu.i2c, MPU6050_RA_GYRO_YOUT_H);
        GYRO_ZOUT_OFFSET_SUM += mpu_read_short(imu.i2c, MPU6050_RA_GYRO_ZOUT_H);

        delay_ms(1);
    }
    imu.gyro_x_offset = GYRO_XOUT_OFFSET_SUM/LOOPS;
    imu.gyro_y_offset = GYRO_YOUT_OFFSET_SUM/LOOPS;
    imu.gyro_z_offset = GYRO_ZOUT_OFFSET_SUM/LOOPS;
}

{int, int, int} get_accel_raw(struct IMU &imu) {
    int ax = mpu_read_short(imu.i2c, MPU6050_RA_ACCEL_XOUT_H);
    int ay = mpu_read_short(imu.i2c, MPU6050_RA_ACCEL_YOUT_H);
    int az = mpu_read_short(imu.i2c, MPU6050_RA_ACCEL_ZOUT_H);

    return {ax, ay, az};

}

{float, float} get_accel_angles(struct IMU &imu) {
    int raw_x, raw_y, raw_z;
    float ax, ay;

    {raw_x, raw_y, raw_z} = get_accel_raw(imu);
    {ax, ay} = norm_accel(raw_x, raw_y, raw_z) ;

    return {ax, ay};
}

float estimate_angle(float gz, float ax, int elapsed) {
    static float prev_angle ;
    static int   is_first_pass = 1;

    float elapsed_sec = (float) elapsed / 100000000;
    float gyro_angle  = gz * elapsed_sec;
    float accel_angle = ax;
    float current;

    if (is_first_pass) {
        current = accel_angle;
        is_first_pass = 0;
    } else {
        current = GYRO_CONTRIBUTION * (prev_angle + gyro_angle) - (1.0 - GYRO_CONTRIBUTION) * accel_angle;
    }
    prev_angle = current;
    return current;
}

#define R2D 57.29577951308232

{float, float} norm_accel(int raw_x, int raw_y, int raw_z) {
    // converts raw accelerometer data into integer 3D euler angles
    float ax ;
    float ay ;

    ax = R2D * atan((float) raw_y / sqrt(pow((float)raw_z,2) + pow((float)raw_x,2)));
    ay = R2D * atan((float)-raw_x / sqrt(pow((float)raw_z,2) + pow((float)raw_y,2)));
    return {ax, ay};
}

