/*
 * pid.xc
 *
 *  Created on: Dec 11, 2012
 *      Author: ah6
 */
#include <math.h>
#include "pid.h"




int pid(pid_state pstate, float input, float dt) {
    float integral_term;
    float error;
    float derivative_term;
    int   correction;

    // compute error terms
    error = pstate.setpoint - input;

    integral_term = pstate.integral_term + (pstate.Ki * error) * dt;
    if(integral_term > pstate.max_correction) integral_term = pstate.max_correction;
    if(integral_term < pstate.min_correction) integral_term = pstate.min_correction;

    derivative_term = (error - pstate.last_error) / dt;

    correction = (pstate.Kp * error) + (integral_term) + (pstate.Kd * derivative_term);

    // hard limit the correction
    if (correction > pstate.max_correction) correction = pstate.max_correction;
    if (correction < pstate.min_correction) correction = pstate.min_correction;

    // save the modified state variables
    pstate.integral_term = integral_term;
    pstate.last_error = error;
    return correction;
}
